@@ -19,11 +19,11 @@ Firewall::PaloAlto - Interact with a Palo Alto firewall's API through Perl.
 
 =head1 VERSION
 
-Version 0.02
+Version 0.03
 
 =cut
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 =head1 SYNOPSIS
 
@@ -57,16 +57,18 @@ if you don't export anything, such as for a purely object-oriented module.
 
 =head1 CLASS METHODS
 
-=head2 connect(%parameters)
+=head2 new(%parameters)
 
-This function connects to the Palo Alto with the credentials specified.
+The constructor generates a new object to interact with a specific firewall.
 
-Host, username and password are mandatory.
-If not specified SSL is used, but it can be disabled using the argument ssl => 0
+The host, username and password parameters are mandatory.
+If not specified, SSL is used, but it can be disabled using the argument ssl => 0
 
 Detailed debugging can be turned on using the debu => 1 argument. It is off by default.
 
-=head3 Arguments
+    my $palo = Firewall::PaloAlto->new(host => 'paloalto.local', username => 'admin', password => 'admin', ssl => 0, debug => 1);
+
+B<Parameters>
 
 =over 
 
@@ -89,23 +91,6 @@ debug (optional, default: 0) - print debugging messages.
 
 =cut
 
-sub connect {
-    my $self = shift; 
-
-    my $user_agent = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });
-    my $key_request = HTTP::Request->new(GET => $self->base_uri.'type=keygen&user='.$self->username.'&password='.$self->password);
-    my $key_response = $user_agent->request($key_request);
-
-    my $pa_response = $self->_check_response_and_parse($key_response);
-
-    if ($pa_response->{status} eq 'success') {
-        $self->_api_key($pa_response->{result}->{key});
-        return 1;
-    }
-    else {
-        return 0;
-    }   
-}
 
 has host => ( is => 'ro',
                 isa => 'Str',
@@ -151,12 +136,13 @@ sub _create_requester {
         my (%request_args) = @_;
         my $user_agent = LWP::UserAgent->new;
 
+        $self->_debug_print("API Request action", $args{action});
+
         for my $key (keys %request_args) {
+            $self->_debug_print("API Request String", "$key=$request_args{$key}");
             $request .= "&$key=$request_args{$key}";
         }
 
-        $self->_debug_print((caller(1))[3], "Sending HTTP Request");
-
         my $http_request = HTTP::Request->new(GET => $request);
         my $http_response = $user_agent->request($http_request);
 
@@ -167,35 +153,6 @@ sub _create_requester {
 sub _check_response_and_parse {
     my $self = shift;
     my $http_response = shift;
-
-    my $response_codes = {
-                            400 => {is_error => 1, string => "Bad Request", parse => sub { "Error 400" } },
-                            403 => { is_error => 1, string => "Forbidden", parse => sub { $_[0]->{result}->{msg}; } },
-                            1 => { is_error => 1, string => "Unknown Command", parse => sub { "Error 1" } },
-                            2 => { is_error => 1, string => "Internal Error (2)", parse => sub { "Error 2" } },
-                            3 => { is_error => 1, string => "Internal Error (3)", parse => sub { "Error 3" } },
-                            4 => { is_error => 1, string => "Internal Error (4)", parse => sub { "Error 4" } },
-                            5 => { is_error => 1, string => "Internal Error (4)", parse => sub { "Error 5" } },
-                            6 => { is_error => 1, string => "Bad XPath", parse => sub { "Error 6" } },
-                            7 => { is_error => 1, string => "Object not present", parse => sub { "Error 7" } },
-                            8 => { is_error => 1, string => "Object not unique", parse => sub { "Error 8" } },
-                            9 => { is_error => 1, string => "Internal Error (9)", parse => sub { "Error 9" } },
-                            10 => { is_error => 1, string => "Reference count not zero", parse => sub { "Error 10" } },
-                            11 => { is_error => 1, string => "Internal Error (11)", parse => sub { "Error 11" } },
-                            12 => { is_error => 1, string => "Invalid Object", parse => sub { join(", ", @{$_[0]->{msg}->{line}}); } },
-                            13 => { is_error => 1, string => "Operation Failed", parse => sub { "Error 13" } },
-                            14 => { is_error => 1, string => "Operation Not Possible", parse => sub { "Error 14" } },
-                            15 => { is_error => 1, string => "Operation Denied", parse => sub { "Error 15" } },
-                            16 => { is_error => 1, string => "Unauthorized", parse => sub { "Error 16" } },
-                            17 => { is_error => 1, string => "Invalid Command", parse => sub { "Error 16" } },
-                            18 => { is_error => 1, string => "Malformed XML", parse => sub { $_[0]->{msg}->{line} } },
-                            19 => { is_error => 0, string => "Get Request Successful", parse => sub{} },
-                            20 => { is_error => 0, string => "Set Request Successful", parse => sub{} },
-                            21 => { is_error => 1, string => "Internal Error (21)", parse => sub { "Error 21" } },
-                            22 => { is_error => 1, string => "Session Timed Out", parse => sub { "Error 22" } },
-                            #Custom code for keygen success
-                            1023 => { is_error => 0, string => "KeyGen Successful", parse => sub {} },
-                        };
    
     #We locally redefine croak so we can get some nice red printing around it.
     no warnings 'redefine';
@@ -203,7 +160,7 @@ sub _check_response_and_parse {
     local *croak = sub {  original_croak("\e[31m".$_[0]."\e[0m"); };
     
     #Check the http response message - croak if its not successful
-    croak "*[HTTP Request] Failed" if !$http_response->is_success;
+    croak "*[HTTP Request] Failed: ".$http_response->content if !$http_response->is_success;
     
     my $palo_response = XMLin($http_response->content, KeyAttr => 'name');
     
@@ -214,16 +171,78 @@ sub _check_response_and_parse {
     }
     
     #Check the response from the Palo Alto:
-    my $code = $response_codes->{$palo_response->{code}};
-    my $api_error = $code->{parse}->($palo_response);
-    croak "*[API Request] $code->{string}: $api_error" if $code->{is_error};
-    
-    $self->_debug_print((caller(1))[3], "API Request $palo_response->{status}");
+    my $code_info = $self->_get_response_info($palo_response->{code});
+    my $api_error = $code_info->{parse}->($palo_response);
+
+    carp "*[API Request] $code_info->{string}: $api_error" and return 0 if $code_info->{is_error};
     
+    $self->_debug_print("API Response", $palo_response->{status});
     
     return $palo_response;
 }
 
+sub _get_response_info {
+    my ($self, $code) = @_;
+
+    my $response_codes = {
+                            400 => {is_error => 1, string => "Bad Request", 
+                                parse => sub { "Error 400" } },
+                            403 => { is_error => 1, string => "Forbidden", 
+                                parse => sub { $_[0]->{result}->{msg}; } 
+                            },
+                            1 => { is_error => 1, string => "Unknown Command", 
+                                parse => sub { "Error 1" } },
+                            2 => { is_error => 1, string => "Internal Error (2)", 
+                                parse => sub { "Error 2" } },
+                            3 => { is_error => 1, string => "Internal Error (3)", 
+                                parse => sub { "Error 3" } },
+                            4 => { is_error => 1, string => "Internal Error (4)", 
+                                parse => sub { "Error 4" } },
+                            5 => { is_error => 1, string => "Internal Error (4)", 
+                                parse => sub { "Error 5" } },
+                            6 => { is_error => 1, string => "Bad XPath", 
+                                parse => sub { "Error 6" } },
+                            7 => { is_error => 1, string => "Object not present", 
+                                parse => sub { "Error 7" } },
+                            8 => { is_error => 1, string => "Object not unique", 
+                                parse => sub { "Error 8" } },
+                            9 => { is_error => 1, string => "Internal Error (9)", 
+                                parse => sub { "Error 9" } },
+                            10 => { is_error => 1, string => "Reference count not zero", 
+                                parse => sub { "Error 10" } },
+                            11 => { is_error => 1, string => "Internal Error (11)", 
+                                parse => sub { "Error 11" } },
+                            12 => { is_error => 1, string => "Invalid Object", 
+                                parse => sub { ref($_[0]->{msg}->{line}) eq 'ARRAY' ? join(" / ", @{ $_[0]->{msg}->{line} }) : $_[0]->{msg}->{line} } 
+                                },
+                            13 => { is_error => 1, string => "Operation Failed", 
+                                parse => sub { $_[0]->{msg}->{line} } },
+                            14 => { is_error => 1, string => "Operation Not Possible", 
+                                parse => sub { "Error 14" } },
+                            15 => { is_error => 1, string => "Operation Denied", 
+                                parse => sub { "Error 15" } },
+                            16 => { is_error => 1, string => "Unauthorized", 
+                                parse => sub { "Error 16" } },
+                            17 => { is_error => 1, string => "Invalid Command", 
+                                parse => sub { "Error 16" } },
+                            18 => { is_error => 1, string => "Malformed XML", 
+                                parse => sub { $_[0]->{msg}->{line} } },
+                            19 => { is_error => 0, string => "Get Request Successful", 
+                                parse => sub{} },
+                            20 => { is_error => 0, string => "Set Request Successful", 
+                                parse => sub{} },
+                            21 => { is_error => 1, string => "Internal Error (21)", 
+                                parse => sub { "Error 21" } },
+                            22 => { is_error => 1, string => "Session Timed Out", 
+                                parse => sub { "Error 22" } },
+                            #Custom code for keygen success
+                            1023 => { is_error => 0, string => "KeyGen Successful", parse => sub {} },
+                        };
+
+    #Return the hash, or 'undef' if the code doesn't exist.
+    return $response_codes->{$code}; 
+}
+
 sub _generate_elements {
     my $self = shift;
     my %element_hash = @_;
@@ -276,13 +295,44 @@ sub _debug_print {
 
 =head1 OBJECT METHODS 
 
+=head2 connect()
+
+The connect functions connects to the Palo Alto, validates and saves the API key.
+It has no parameters
+
+ $pa->connect();
+
+=cut
+
+sub connect {
+    my $self = shift; 
+
+    my $user_agent = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });
+    my $key_request = HTTP::Request->new(GET => $self->base_uri.'type=keygen&user='.$self->username.'&password='.$self->password);
+
+    $self->_debug_print("Key Request", "");
+
+    my $key_response = $user_agent->request($key_request);
+
+    my $pa_response = $self->_check_response_and_parse($key_response);
+
+    if ($pa_response->{status} eq 'success') {
+        $self->_api_key($pa_response->{result}->{key});
+        return 1;
+    }
+    else {
+        return 0;
+    }   
+}
+
+
 =head2 commit(%parameters)
 
 The commit function commits the current candidate configuration to the Palo Alto firewall.
 
     $pa_firewall->commit(vsys_id => 10, poll => 10);
 
-=head3 Arguments
+B<Parameters>
 
 =over 
 
@@ -326,7 +376,11 @@ sub commit {
 
 =head2 vsys($action, %parameters)
 
-=head3 Arguments
+The vsys function creates a new virtual system.
+
+ $pa->vsys('set', vsys_id => 4, display-name => "Tenant 4");
+
+B<Parameters>
 
 =over
 
@@ -363,9 +417,11 @@ sub vsys {
 
 =head2 vsys_import($action, %parameters);
 
+The vsys_import function imports interfaces and virtual-routers into a virtual system. 
+
     $pa_firewall->vsys_import('set', vsys_id => 4, interface => [ ethernet1/1, ethernet1/2 ], virtual-router => [ default ]);
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -405,7 +461,11 @@ sub vsys_import {
 
 =head2 l3_subinterface($action, %parameters)
 
-=head3 Arguments
+This function creates a new layer 3 subinterface underneath a parent interface.
+
+    $pa->l3_subinterface('set', parent => 'ethernet1/1', tag => 5, description => 'Tenant x untrust interface');
+
+B<Parameters>
 
 =over
 
@@ -440,7 +500,7 @@ sub l3_subinterface {
 
 =head2 virtual_router($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -471,7 +531,7 @@ sub virtual_router {
 
 =head2 zone($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -503,7 +563,7 @@ sub zone {
 
 =head2 ipv4_static_route($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -542,7 +602,7 @@ sub ipv4_static_route {
 
 =head2 ipv6_static_route($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -581,7 +641,7 @@ sub ipv6_static_route {
 
 =head2 address($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -613,13 +673,13 @@ sub address {
 
     my $elements = $self->_generate_elements(%args);
 
-    return $requester->(xpath => "/config/devices/entry[\@name='localhost.localdomain']/vsys/entry[\@name=\'$vsys\']/address/entry[$address]/static", element => $elements);
+    return $requester->(xpath => "/config/devices/entry[\@name='localhost.localdomain']/vsys/entry[\@name=\'$vsys\']/address/entry[$address]", element => $elements);
 }
 
 
 =head2 address_group($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
@@ -655,9 +715,11 @@ sub address_group {
     return $requester->(xpath => "/config/devices/entry/vsys/entry[\@name=\'$vsys\']/address-group/entry[$name]", element => $elements);
 }
 
+
+
 =head2 security_rule($action, %parameters)
 
-=head3 Arguments
+B<Parameters>
 
 =over
 
